package tp4509.picross;


import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Pattern;


/**
 * Classe qui gére une grille picross (proposition ou solution).
 * @author Eric Lallet.
 *
 */
public class GrillePicross {
	/**
	 * Valeur pour les cases d'indices sans usage.
	 */
	public static final int SANSINDICE = -1;
	/**
	 * nombre de colonnes dans la grille.
	 */
	private int largeur;
	/**
	 * nombre de lignes dans la grille.
	 */
	private int hauteur;
	
	
	/**
	 * valeurs de la grilles.
	 */
	private int[][] grille;
	/**
	 * liste des informations pour chaque ligne.
	 */
	private List<Integer>[] indicesLigne;
	/**
	 * liste des informations pour chaque colonne.
	 */
	private List<Integer>[]  indicesColonne;
	
	/**
	 * Création d'une grille largeurXhauteur vide.
	 * @param largeur
	 *          nombre de colonnes de la grille. 
	 * @param hauteur
	 *          nombre de lignes de la grille.
	 */
	public GrillePicross(final int largeur, final int hauteur) {
		this.largeur = largeur;
		this.hauteur = hauteur;
		grille = new int[largeur][hauteur];
		creerIndices();
		remplirIndices();
	}
	
	/**
	 * Création d'une grille conforme aux données du fichier fileName.
	 * @param fileName
	 *           nom du fichier de données sur la grille.
	 * @throws IOException
	 *           exceptions sur l'ouverture ou la lectur du fichier.
	 * @throws FileFormatException
	 *           exceptions sur le format du contenu du fichier.
	 */
	public GrillePicross(final String fileName) throws IOException, FileFormatException {
		chargerGrille(fileName);
		creerIndices();
		remplirIndices();
	}
	
	/**
	 * Chargement d'une grille à partir du fichier fileName.
	 * @param fileName
	 *           nom du fichier de données sur la grille.
	 * @throws IOException
	 *           exceptions sur l'ouverture ou la lectur du fichier.
	 * @throws FileFormatException
	 *           exceptions sur le format du contenu du fichier.
	 */
	public void chargerGrille(final String fileName) throws IOException, FileFormatException {
		//Pattern patternePremiereLigne = Pattern.compile("^\\d+x\\d+$"); // une ligne avec: nbxnb
		BufferedReader br = Files.newBufferedReader(Paths.get(fileName));
		String premiereLigne = br.readLine();
		String ligne;
		if (!Pattern.matches("^\\d+x\\d+$", premiereLigne)) {
			throw new FileFormatException("Première ligne du fichier " + fileName + " incorrecte: " + premiereLigne);
		}
		Pattern paternSeparateur = Pattern.compile("x");
		String[] numbers = paternSeparateur.split(premiereLigne);
		this.largeur = Integer.parseInt(numbers[0]);
		this.hauteur = Integer.parseInt(numbers[1]);
		this.grille = new int[largeur][hauteur];
		for (int lig = 0; lig < hauteur; lig++) {
			ligne = br.readLine();
			if (ligne == null) {
				throw new FileFormatException(fileName + ":" + lig + " lignes présentes sur " + hauteur + " lignes attendues.");
			}
			for (int col = 0; col < largeur; col++) {
				char car = ligne.charAt(col);
				grille[col][lig] = PicrossColor.colorByChar(car).ordinal();
			}
		}
	}
	
	/**
	 * Crée les tableaux des listes des indices pour les lignes et les colonnes. 
	 * Les listes sont vides.
	 */
	private void creerIndices() {
		indicesLigne = new List[hauteur];
		for (int lig = 0; lig < hauteur; lig++) {
			indicesLigne[lig] = new ArrayList<Integer>();
		}
		indicesColonne = new List[largeur];
		for (int col = 0; col < largeur; col++) {
			indicesColonne[col] = new ArrayList<Integer>();
		}
	}
	
	/**
	 * Rempli la liste des indices pour la ligne numéro lig.
	 * Version pour picross noir/blanc.
	 * @param lig
	 *        numéro de la ligne dont on remplit les indices.
	 */
	private void remplirIndicesLigne(final int lig) {
		indicesLigne[lig].clear();
		int noneValue = PicrossColor.NONE.ordinal();
		int seq  = 0;
		for (int col = 0; col < largeur; col++) {
			if (grille[col][lig] != noneValue) {
				seq++;
			} else {
				if (seq > 0) {
					indicesLigne[lig].add(seq);
				}
				seq = 0;
			}
		}
		if (seq > 0) {
			indicesLigne[lig].add(seq);
		}
		if (indicesLigne[lig].isEmpty()) {
			indicesLigne[lig].add(0);
		}
	}
	
	/**
	 * Rempli la liste des indices pour la colonne numéro col.
	 * @param col
	 *        numéro de la colonne dont on remplit les indices.
	 */
	private void remplirIndicesColonne(final int col) {
		indicesColonne[col].clear();
		int noneValue = PicrossColor.NONE.ordinal();
		int seq  = 0;
		for (int lig = 0; lig < hauteur; lig++) {
			if (grille[col][lig] != noneValue) {
				seq++;
			} else {
				if (seq > 0) {
					indicesColonne[col].add(seq);
				}
				seq = 0;
			}
		}
		if (seq > 0) {
			indicesColonne[col].add(seq);
		}
		if (indicesColonne[col].isEmpty()) {
			indicesColonne[col].add(0);
		}
	}
	
	/**
	 * remplit les indices de toutes les colonnes et de toutes les lignes.
	 */
	private void remplirIndices() {
		for (int col = 0; col < largeur; col++) {
			remplirIndicesColonne(col);
		} 
		
		for (int lig = 0; lig < hauteur; lig++) {
			remplirIndicesLigne(lig);
		}
		
		
		
	}
	
	/**
	 * affiche les indices de toutes les lignes et toutes les colonnes.
	 */
	public void afficherIndices() {
		for (int lig = 0; lig < hauteur; lig++) {
			for (Integer val: indicesLigne[lig]) {
				System.out.print(val + " ");
			}
			System.out.println();
		}
		
		for (int col = 0; col < largeur; col++) {
			for (Integer val: indicesColonne[col]) {
				System.out.print(val + " ");
			}
			System.out.println();
		}
	}
	
	/**
	 * retourne le nombre de colonnes de la grilles.
	 * @return
	 *       le nombre de colonnes de la grille.
	 */
	public int getLargeur() {
		return largeur;
	}

	/**
	 * retourne le nombre de ligne de la grille.
	 * @return
	 *       le nombre de lignes de la grille.
	 */
	public int getHauteur() {
		return hauteur;
	}
	
	/**
	 * Fixe une valeur (false ou true)  pour la case colXlig de la grille et modifie les indices en conséquence.
	 * @param col
	 *          colonne de la case modifiée.
	 * @param lig
	 *          ligne de la case modifiée.
	 * @param value
	 *          valeur placée dans la case.
	 * @throws IllegalArgumentException
	 *          exception si les numéros de la case sont en dehors de la grille.
	 */
	public void fixerCase(final int col, final int lig, final int value) throws IllegalArgumentException {
		if ((lig < 0) || (lig >= hauteur) || (col < 0) || (col >= largeur)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors de la grille " +  largeur + "x" + hauteur);
		}
		grille[col][lig] = value;
		remplirIndicesColonne(col);
		remplirIndicesLigne(lig);
	}
	
	/**
	 * Fixe à BLACK  la case colXlig de la grille et modifie les indices en conséquence.
	 * @param col
	 *          colonne de la case modifiée.
	 * @param lig
	 *          ligne de la case modifiée.
	 * @throws IllegalArgumentException
	 *          exception si les numéros de la case sont en dehors de la grille.
	 */
	public void marquerCase(final int col, final int lig) throws IllegalArgumentException {
		fixerCase(col, lig, PicrossColor.BLACK.ordinal());
	}
	
	/**
	 * Indique si une case de la grille a été noircie.
	 * @param col
	 *        colonne de la case.
	 * @param lig
	 *        ligne de calse.
	 * @return
	 *        true si la case est noircie, false sinon.
	 * @throws IllegalArgumentException
	 *         Exception levée si les coordonnées de la case sont en dehors de la grille.
	 */
	public boolean estNoir(final int col, final int lig) throws IllegalArgumentException {
		if ((lig < 0) || (lig >= hauteur) || (col < 0) || (col >= largeur)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors de la grille " +  largeur + "x" + hauteur);
		}
		return grille[col][lig] == PicrossColor.BLACK.ordinal();
		
	}
	/**
	 * Fixe à NONE  la case colXlig de la grille et modifie les indices en conséquence.
	 * @param col
	 *          colonne de la case modifiée.
	 * @param lig
	 *          ligne de la case modifiée.
	 * @throws IllegalArgumentException
	 *          exception si les numéros de la case sont en dehors de la grille.
	 */
	public void effacerCase(final int col, final int lig) throws IllegalArgumentException {
		fixerCase(col, lig, PicrossColor.NONE.ordinal());
	}
	
	/**
	 * Inverse la valeur de la case colXlig de la grille et modifie les indices en conséquence.
	 * @param col
	 *          colonne de la case modifiée.
	 * @param lig
	 *          ligne de la case modifiée.
	 * @throws IllegalArgumentException
	 *          exception si les numéros de la case sont en dehors de la grille.
	 */
	public void changerCase(final int col, final int lig)  throws IllegalArgumentException {
		if ((lig < 0) || (lig >= hauteur) || (col < 0) || (col >= largeur)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors de la grille " +  largeur + "x" + hauteur);
		}
		if (grille[col][lig] == PicrossColor.BLACK.ordinal()) {
			effacerCase(col, lig);
		} else {
			marquerCase(col, lig);
		}
	}
	
	/**
	 * Calcule la taille de la plus grande liste dans un tableau de liste.
	 * @param listesTab
	 *        le tableau de liste.
	 * @return
	 *         la taille de la plus grande liste dans le tableau de liste.
	 */
	private int getTailleMaxList(final List<Integer>[] listesTab) {
		int max = 0;
		for (List<Integer> liste: listesTab) {
			if (max < liste.size()) {
				max = liste.size();
			}
		}
		return max;
	}
	
	/**
	 * Calcule la largeur de la zone des indices sur les lignes.
	 * @return
	 *        la largeur de la zone des indices sur les lignes.
	 */
	public int getLargeurIndices() {
		return getTailleMaxList(indicesLigne);
	}
	
	/**
	 *  Calcule la hauteur de la zone des indices sur les colonnes.
	 * @return
	 *        la hauteur de la zone des indices sur les colonnes.
	 */
	public int getHauteurIndices() {
		return getTailleMaxList(indicesColonne);
	}
	

	/**
	 * Donne la valeur de case col x lig de la grille de jeu.
	 * @param col
	 *        colonne de la case.
	 * @param lig
	 *        ligne de la case.
	 * @return
	 *       valeur de la case.
	 */
	public int getValeur(final int col, final int lig) {
		if ((lig < 0) || (lig >= hauteur) || (col < 0) || (col >= largeur)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors de la grille " +  largeur + "x" + hauteur);
		}
		return grille[col][lig];
	}

	/**
	 * Donne la valeur de l'indice placée à col x lig sans la zone des indices sur les colonnes.
	 * @param col
	 *        colonne dans la zone d'indices.
	 * @param lig
	 *        ligne dans la zone d'indice.
	 * @return
	 *        la valeur de l'indice.
	 */
	public int getIndiceColonne(final int col, final int lig) {
		int hauteurIndices = getHauteurIndices();
		if ((lig < 0) || (lig >= hauteurIndices) || (col < 0) || (col >= largeur)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors des indices " +  largeur + "x" + hauteurIndices);
		}
		int tailleIndice = indicesColonne[col].size();
		if (lig < (hauteurIndices - tailleIndice)) {
			return SANSINDICE;
		}
		return indicesColonne[col].get(lig - hauteurIndices + tailleIndice);	
	}
	
	/**
	 * Donne la valeur de l'indice placée à col x lig sans la zone des indices sur les lignes.
	 * @param col
	 *        colonne dans la zone d'indices.
	 * @param lig
	 *        ligne dans la zone d'indice.
	 * @return
	 *        la valeur de l'indice.
	 */
	public int getIndiceLigne(final int col, final int lig) {
		int largeurIndices = getLargeurIndices();
		if ((lig < 0) || (lig >= hauteur) || (col < 0) || (col >= largeurIndices)) {
			throw new IllegalArgumentException("Le point " + col + "," + lig + " est en dehors des indices " +  largeurIndices + "x" + hauteur);
		}
		int tailleIndice = indicesLigne[lig].size();
		if (col < (largeurIndices - tailleIndice)) {
			return SANSINDICE;
		}
		return indicesLigne[lig].get(col - largeurIndices + tailleIndice);	
	}
	
	
	//CHECKSTYLE:OFF
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(indicesColonne);
		result = prime * result + Arrays.hashCode(indicesLigne);
		return result;
	}

	// deux grilles sont égales si elles génères les mêmes indices.
	// en cas de grilles ambigües, toute proposition conforme aux indices
	// sera considérée comme correcte.
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		GrillePicross other = (GrillePicross) obj;
		if (!Arrays.equals(indicesColonne, other.indicesColonne))
			return false;
		if (!Arrays.equals(indicesLigne, other.indicesLigne))
			return false;
		return true;
	}

	
		
	@Override
	public String toString() {
		String str = hauteur + "x" + largeur + "\n";
		for (int lig = 0; lig < hauteur; lig++) {
			for(int col = 0; col < largeur; col++) {
				str += PicrossColor.values()[grille[col][lig]].getCar();
			}
			str += "\n";
		}
		return str;
	}
		

}
