package tp4509.tcp;



import static dm4509.common.Log.COMM;
import static dm4509.common.Log.LOG_ON;

import java.nio.ByteBuffer;

import org.apache.log4j.Level;

/**
 * Classe qui construit les messages du protocole sous leurs deux formes: les 4 champs, où le buffer qui contient la trame à
 * transmettre sur le réseau.
 * @author Eric Lallet.
 *
 */
public class MessageProtocole {
	

	/**
	 * Nombre d'entiers dans les trames échangées.
	 */
	private static final int NBINTEGERS = 4;
	
	/**
	 * taille du buffer des trames échangées.
	 */
	public static final int BUFFERSIZE =  NBINTEGERS * (Integer.SIZE / Byte.SIZE);  
	
	/**
	 * type du message  (dont la traduction en entier doit être placée en tête du buffer). 
	 */
	private Protocole type;
	
	/**
	 * l'entier mis en deuxième position dans la trame.
	 */
	private int val1;
	
	/**
	 * l'entier mis en troisime position dans la trame.
	 */
	private int val2;
	
	/**
	 * l'entier mis en quatrième position dans la trame.
	 */
	private int val3;
	
	
	/**
	 * buffer contenant les quatre entier à envoyer. 
	 */
	private ByteBuffer buffer;
	
	
	/**
	 * Construction à partir du type et des trois entiers.
	 * @param type
	 *            type du message (dont la traduction en entier doit être placée en tête du buffer). 
	 * @param val1
	 *            deuxième entier à placer dans le buffer.
	 * @param val2
	 *            troisième entier à placer dans le buffer.
	 * @param val3
	 *            quatrième entier à placer dans le buffer.
	 */
	public MessageProtocole(final Protocole type, final int val1, final int val2, final int val3) {
		buffer = ByteBuffer.allocate(BUFFERSIZE);
		buffer.putInt(type.ordinal());
		buffer.putInt(val1);
		buffer.putInt(val2);
		buffer.putInt(val3);
		buffer.flip();
		this.type = type;
		this.val1 = val1;
		this.val2 = val2;
		this.val3 = val3;
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[MessageProtcole::MessageProtocole] Contructions de " + type + "(" + val1 + ", " + val2 +  ", " + val3 + ").");
		}
	}
	
	/**
	 * Construction à partir d'un buffer  rempli de quatre entiers prêt pour la lecture.
	 * @param buffer
	 *          buffer rempli de quatre entiers prêt pour la lecture.
	 */
	public MessageProtocole(final ByteBuffer buffer) {
		this.buffer = buffer;
		int typeInt = buffer.getInt();
		type = Protocole.values()[typeInt];
		val1 = buffer.getInt();
		val2 = buffer.getInt();
		val3 = buffer.getInt();
		
		buffer.rewind();
	}
	
	

	/**
	 * getter sur le type du message.
	 * @return
	 *         type du message  (dont la traduction en entier doit être placée en tête du buffer).
	 */
	public Protocole getType() {
		return type;
	}


	/**
	 * getter sur le deuxième entier de la trame.
	 * @return
	 *        deuxième entier à placer dans le buffer.
	 */
	public int getVal1() {
		return val1;
	}

	/**
	 * getter sur le troisième entier de la trame.
	 * @return
	 *        troisième entier à placer dans le buffer.
	 */
	public int getVal2() {
		return val2;
	}


	/**
	 * getter sur le quatrième entier de la trame.
	 * @return
	 *        quatrième entier à placer dans le buffer.
	 */
	public int getVal3() {
		return val3;
	}

	/**
	 * getter sur le buffer prêt à l'envoi, contenant les quatres entier d'une trame.
	 * @return
	 *       le buffer prêt à l'envoi, contenant les quatres entier d'une trame.
	 */
	public ByteBuffer getBuffer() {
		buffer.rewind(); // On remet le buffer prêt pour une lectures des données.
		return buffer;
	}
	
	@Override
	public String toString() {
		return type + "(" + val1 + ", " + val2 + ", " + val3 + ")";
	}

	//CHECKSTYLE:OFF
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((type == null) ? 0 : type.hashCode());
		result = prime * result + val1;
		result = prime * result + val2;
		result = prime * result + val3;
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		MessageProtocole other = (MessageProtocole) obj;
		if (type != other.type)
			return false;
		if (val1 != other.val1)
			return false;
		if (val2 != other.val2)
			return false;
		if (val3 != other.val3)
			return false;
		return true;
	}
	
}
